# -*- coding: utf-8 -*-

#  docprop.py - Python plugin to display document properties
#  
#  Copyright (C) 2006  Sayamindu Dasgupta
#  Copyright (C) 2005-2006 Igalia
#  Copyright (C) 2006 Matthew Dugan
#  
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#   
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#   
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 59 Temple Place, Suite 330,
#  Boston, MA 02111-1307, USA.

import datetime
import grp
import gtk
import gtk.glade
import gnomevfs
import gedit
import pango
import pwd
import os
from gettext import gettext as _

properties_str = """
<ui>
  <menubar name="MenuBar">
    <menu name="FileMenu" action="File">
      <separator/>
      <placeholder name="FileOps_4">
            <menuitem name="Properties" action="ShowPropWin"/>
      </placeholder>
    </menu>
  </menubar>
</ui>
"""

class DocpropPlugin(gedit.Plugin):
    WINDOW_DATA_KEY = "DocpropPluginWindowData"
    
    def __init__(self):
        gedit.Plugin.__init__(self)

    def activate(self, window):
        add_action_group(self, window)
        test_activation_group(window)          
    
    def deactivate(self, window):
        remove_action_group(window)
        
    def update_ui(self, window):
        test_activation_group(window)

    def show_properties(self, action, window):
        tab = window.get_active_tab()
        if tab is not None:
            document = tab.get_document()
            doc_type = document.get_mime_type()
            doc_uri = document.get_uri()
            
            doc_info = gnomevfs.get_file_info(doc_uri, gnomevfs.FILE_INFO_DEFAULT)
            
            gladefile = os.path.join(os.path.dirname(__file__), 'docprop.glade2')
            wTree = gtk.glade.XML(gladefile)
            
            win = wTree.get_widget("dialog")
            
            label = wTree.get_widget("file_name_label")
            text =  '<span weight="bold" size="larger">' + document.get_short_name_for_display() + '</span>'
            label.set_markup(text)
            
            label = wTree.get_widget("label_filetype")
            text =  gnomevfs.mime_get_description(doc_type)
            label.set_text(text)
            
            label = wTree.get_widget("label_filesize")
            if doc_info.size > 1048576: 
                text = '%.2f MB' %(doc_info.size/1048576.0)
                text = text + " ("+str(doc_info.size)+_(" bytes")+")"
            elif doc_info.size > 1024: 
                text = '%.2f KB' %(doc_info.size/1024.0)
                text = text + " ("+str(doc_info.size)+_(" bytes")+")"                
            else:
                text = str(doc_info.size)+_(" bytes")
            label.set_text(text)

            label = wTree.get_widget("label_location")
            if doc_info.flags == gnomevfs.FILE_FLAGS_LOCAL:
                desc = doc_uri[7:]
            else:
                desc = doc_uri
            label.set_ellipsize(pango.ELLIPSIZE_END)
            label.set_text(desc)

            
            label = wTree.get_widget("label_mimetype")
            label.set_text(doc_type)            
  
            label = wTree.get_widget("label_owner")
            if doc_info.flags == gnomevfs.FILE_FLAGS_LOCAL:
                desc = pwd.getpwuid(doc_info.uid)[0]
            else:
                desc = str(doc_info.uid)
            label.set_text(desc)

            label = wTree.get_widget("label_group")
            if doc_info.flags == gnomevfs.FILE_FLAGS_LOCAL:
                desc = grp.getgrgid(doc_info.gid)[0]
            else:
                desc = str(doc_info.gid)
            label.set_text(desc)

            label = wTree.get_widget("label_permissions")
            # Figure out the permissions string. There *must* be a saner way of doing this
            desc = '-'
            if doc_info.permissions & gnomevfs.PERM_USER_READ:
                desc = desc + 'r'
            else:
                desc = desc + '-'
            if doc_info.permissions & gnomevfs.PERM_USER_WRITE:
                desc = desc + 'w'
            else:
                desc = desc + '-'
            if doc_info.permissions & gnomevfs.PERM_USER_EXEC:
                desc = desc + 'x'
            elif doc_info.permissions & gnomevfs.PERM_SUID:
                desc = desc + 's'
            else:
                desc = desc + '-'
            if doc_info.permissions & gnomevfs.PERM_GROUP_READ:
                desc = desc + 'r'
            else:
                desc = desc + '-'
            if doc_info.permissions & gnomevfs.PERM_GROUP_WRITE:
                desc = desc + 'w'
            else:
                desc = desc + '-'
            if doc_info.permissions & gnomevfs.PERM_GROUP_EXEC:
                desc = desc + 'x'
            elif doc_info.permissions & gnomevfs.PERM_SGID:
                desc = desc + 's'                
            else:
                desc = desc + '-'
            if doc_info.permissions & gnomevfs.PERM_OTHER_READ:
                desc = desc + 'r'
            else:
                desc = desc + '-'
            if doc_info.permissions & gnomevfs.PERM_OTHER_WRITE:
                desc = desc + 'w'
            else:
                desc = desc + '-'
            if doc_info.permissions & gnomevfs.PERM_OTHER_EXEC:
                desc = desc + 'x'
            elif doc_info.permissions & gnomevfs.PERM_STICKY:
                desc = desc + 't'                
            else:
                desc = desc + '-'
            label.set_text(desc)
  
            label = wTree.get_widget("label_modified")
            timestamp = doc_info.mtime
            modtime = datetime.datetime.fromtimestamp(timestamp)
            desc = modtime.strftime("%a %d %b %Y %I:%M:%S %p %Z")
            label.set_text(desc)

            label = wTree.get_widget("label_accessed")
            timestamp = doc_info.atime
            modtime = datetime.datetime.fromtimestamp(timestamp)
            desc = modtime.strftime("%a %d %b %Y %I:%M:%S %p %Z")
            label.set_text(desc)
            
            close_button = wTree.get_widget("close_button")
            close_button.connect_object("clicked", gtk.Widget.destroy, win)
           
            win.show()

        

def add_action_group(self, window):
    manager = window.get_ui_manager()
    
    windowdata = dict()

    windowdata["action_group"] = gtk.ActionGroup(
        "GeditFilePropertiesPluginActions")
    windowdata["action_group"].add_actions(
        [('ShowPropWin', None, _('_Properties'),
          "<control>r",
          _("Show file properties"),
          self.show_properties)],
        window)
    
    manager.insert_action_group(windowdata["action_group"], 0)
    windowdata["ui_id"] = manager.add_ui_from_string(properties_str)
    
    window.set_data("FilePropertiesPluginInfo", windowdata)

def test_activation_group(window):
    manager = window.get_ui_manager()        
    widget = manager.get_widget("/MenuBar/FileMenu/FileOps_4/Properties")
    tab = window.get_active_tab()
    if tab is not None:
        document = tab.get_document()
        doc_uri = document.get_uri()
        if doc_uri is None:
            widget.set_sensitive(False)
        else:
            widget.set_sensitive(True)            
    else:   
        widget.set_sensitive(False)
    
def remove_action_group(window):
    windowdata = window.get_data("FilePropertiesPluginInfo")        
    manager = window.get_ui_manager()        
    manager.remove_ui(windowdata["ui_id"])
    manager.remove_action_group(windowdata["action_group"])
    manager.ensure_update()

# ex:ts=4:et:
